;;; $Id: port-bigloo.scm,v 1.10 2000/10/29 16:36:13 queinnec Exp $
;;; Copyright (c) 1990-98 by Christian Queinnec. All rights reserved.
;;; This file is part of LiSP2TeX, a LiSP and TeX merger.
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;;                                                    LiSP2TeX
;;;   Christian Queinnec
;;;   <Christian.Queinnec@lip6.fr> 
;;;   LIP6 --- Universit Paris 6
;;;   4, place Jussieu
;;;   75252 Paris Cedex 05
;;;   France              
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;;    This program is free software; you can redistribute it and/or modify
;;;    it under the terms of the GNU General Public License as published by
;;;    the Free Software Foundation; either version 2 of the License, or
;;;    (at your option) any later version.
;;;
;;;    This program is distributed in the hope that it will be useful,
;;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;    GNU General Public License for more details.
;;;
;;;    You should have received a copy of the GNU General Public License
;;;    along with this program; if not, write to the Free Software Foundation,
;;;    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; This file contains the functions that are highly dependent of the
;;; Scheme implementation and, here, Bigloo from Manuel Serrano.
;;; NOTE: You need at least Bigloo 1.4 to run correctly LiSP2TeX.
;;; It also contains some utilities that are not present.

;;; The name of the underlying Scheme system.

(define scm-name "Bigloo")

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Utilities are mapcan, atom?, string-uppercase and
;;; string-lowercase. Some ports do not need to define them as they
;;; may already be present.

(define (mapcan fn l)
  (if (pair? l)
      (append (fn (car l))
              (mapcan fn (cdr l)) )
      '() ) )

(define (atom? e)
  (not (pair? e)) )

(define (string-uppercase string)
  (let* ((n (string-length string))
         (s (make-string n)) )
    (do ((i 0 (+ i 1)))
        ((= i n) s)
      (string-set! s i (char-upcase (string-ref string i))) ) ) )

(define (string-lowercase string)
  (let* ((n (string-length string))
         (s (make-string n)) )
    (do ((i 0 (+ i 1)))
        ((= i n) s)
      (string-set! s i (char-downcase (string-ref string i))) ) ) )

;;;ooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Bigloo dependencies.

;;; Only used in their dynamic extent so bind-exit is safe.
;;; Unfortunately, I now need a real call/cc for some examples
;;; so I compile with -call/cc now.

;(define (call/cc f)
;  (bind-exit (k) (f k)) )

;;; Bigloo has a working peek-char.

(define system-can-peek-char? #t)

;;; The three POSIX streams.

(define stdin-port (current-input-port))
(define stdout-port (current-output-port))
(define stderr-port (current-error-port))

;;; Defines case sensitivity.

(define *respect-input* #t)
(define *uppercase-input* #f)
(define *lowercase-input* #f)

;;; Set how expressions to be evaluated (by Bigloo) should be read.
;;; Bigloo reads symbols in uppercase.

(define *evaluator-preferred-respect-input*   #f)
(define *evaluator-preferred-uppercase-input* #t)
(define *evaluator-preferred-lowercase-input* #f)

;;; By default respect case sensitivity.

(define *respect-output* #t)
(define *uppercase-output* #f)
(define *lowercase-output* #f)

;;; After each directive processing, flush buffers if not automatically
;;; provided by the implementation. 

(define (flush-all-buffers) 
  (flush-output-port stdout-port)
  (flush-output-port stderr-port ) )

;;; probe-file tests if a file exists.

(define probe-file file-exists?)

;;; Dynamically evaluates an expression. If an error occurs, wrap the
;;; error in a list with a distinguished symbol.
;;; This is specific to Bigloo.

(define (LiSP2TeX-eval exp)
  (try (eval exp)
       (lambda (k a b c)
         (if *stop-at-first-error*
             (begin 
               (notify-error a b c)
               (exit 1) )
             (k (list '***EVALUATION-ERROR*** a b c)) ) ) ) )

;;; exit prematurely

(define (LiSP2TeX-exit status)
  (flush-all-buffers)
  (exit status) )

;;; call unix

(define (LiSP2TeX-system cmd)
  (system cmd) )

;;; The entry point of LiSP2TeX. The args variable is the list of 
;;; arguments of the shell invokation (a list of strings). This function
;;; has to be called as the initial function.

(define (unix-main args)
  (set! args (cdr args)) ; strip executable name
  (LiSP2TeX args) )

;;; load the content of file. Load it non verbosely to not clutter the
;;; output stream.  The content of the file may use the various macros
;;; that are defined in emit.scm. It is better to load the file not in
;;; verbose mode to avoid cluttering the output: loadq does this.

(define (load-parameter-file file)
  (loadq file) )

;;; Define macros. They must be present for the compilation and the
;;; run-time since they are necessary to load customization files.
;;; New macros can be defined in customization files if they are defined
;;; with define-lisp2tex-macro.

(define-macro (define-lisp2tex-macro call . body)
  `(begin 
     (eval '(define-macro ,call . ,body))
     (define-macro ,call . ,body) ) )

;;; Make define-lisp2tex-macro pervasive itself  

(eval '(define-macro (define-lisp2tex-macro call . body)
         `(define-macro ,call . ,body) ))

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; Define some macros

;;; PC Scheme has still a binary apply so any port has to define this macro.
;;; But this macro is only needed for compilation.

(define-lisp2tex-macro (LiSP2TeX-apply . args)
  `(apply . ,args) )

(define-lisp2tex-macro (unless condition . body)
  `(if ,condition #f (begin . ,body)) )

(define-lisp2tex-macro (when condition . body)
  `(if ,condition (begin . ,body)) )

;;; end of port-bigloo.scm
