;;; $Id: scan-file.scm,v 1.10 2000/10/29 16:36:13 queinnec Exp $
;;; Copyright (c) 1990-98 by Christian Queinnec. All rights reserved.
;;; This file is part of LiSP2TeX, a LiSP and TeX merger.
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;;                                                    LiSP2TeX
;;;   Christian Queinnec
;;;   <Christian.Queinnec@lip6.fr> 
;;;   LIP6 --- Universit Paris 6
;;;   4, place Jussieu
;;;   75252 Paris Cedex 05
;;;   France              
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;;    This program is free software; you can redistribute it and/or modify
;;;    it under the terms of the GNU General Public License as published by
;;;    the Free Software Foundation; either version 2 of the License, or
;;;    (at your option) any later version.
;;;
;;;    This program is distributed in the hope that it will be useful,
;;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;    GNU General Public License for more details.
;;;
;;;    You should have received a copy of the GNU General Public License
;;;    along with this program; if not, write to the Free Software Foundation,
;;;    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;; This file contains the functions ro read a file from disk, scan it
;;; to find the bounds of Sexpressions and cache it in memory.

(define (file->string filename)
  (call-with-input-file filename
    (lambda (in)
      (let ((extensible-string (make-extensible-string 4096)))
	;; read the whole file into an extensible-string.
	;; This code is completely inefficient. It should be better
	;; to determine the size of the file and to read it in one go.
	(letrec ((loop 
		  (lambda (char)
		    (cond 
		     ((eof-object? char) )
		     (else 
		      (increment-string char extensible-string)
		      (loop (read-char in)) ) ) ) ))
	  (loop (read-char in)) )
        (finalize-string extensible-string) ) ) ) )  

;;; convert a file into a scanned-file object
(define (scan-file filename)
  (call-with-input-file filename
    (lambda (in)
      (let ((extensible-string (make-extensible-string 4096))
	    (eof-object 'wait) )
	;; read the whole file into an extensible-string.
	;; This code is completely inefficient. It should be better
	;; to determine the size of the file and to read it in one go.
	(letrec ((loop 
		  (lambda (char)
		    (cond 
		     ((eof-object? char)
		      (set! eof-object char) )
		     (else 
		      (increment-string char extensible-string)
		      (loop (read-char in)) ) ) ) ))
	  (loop (read-char in)) )
	;; determine the bounds of each Sexpression
	(let* ((string (finalize-string extensible-string))
	       (bounds
		(scan-string 
		 (make-counting-stream 
		  filename
		  string
		  (extensible-string-index extensible-string)
		  0 
		  eof-object 
		  0 )
		 (list) ) ) )
	  ;; build the final object
	  (make-scanned-file
	   filename
	   (apply vector (reverse! bounds))
	   string ) ) ) ) ) )

;;; determine the bounds of each Sexpression in the stream.
;;; The substring between start and real-start may contain useful comments
;;; but leading spaces are omitted.
(define (scan-string counting-stream result)
  (let* ((start 
          (begin (counting-skip-spaces counting-stream)
                 (counting-stream-position counting-stream) ) )
         (real-start 
          (begin (counting-skip counting-stream)
                 (counting-stream-position counting-stream) ) )
         (o (counting-read counting-stream)) )
    (if (eof-object? o)
        result
        (if (atom? o)
            (scan-string counting-stream result)
            (scan-string 
             counting-stream
             (cons (make-scanned-expression
                    o
                    (find-scanned-expression-type o)
                    (find-scanned-expression-key o)
                    start
                    real-start
                    (counting-stream-position counting-stream) )
                   result ) ) ) ) ) )

;;; To ease the simple case where an expression is looked for from 
;;; its name precompute the name and the type (ie the definer) 
;;; of the expression

;;; The type is usually define, defmacro ...
(define (find-scanned-expression-type o)
  (if (symbol? (car o))
      (cond ((symbol? (car o)) (symbol->string (car o)))
            ((string? (car o)) (car o))
            (else #f) )
      #f ) )

;;; the defined name generally appears as the first symbol (or atom)
;;; after the type of the expression.
(define (find-scanned-expression-key o)
  (if (pair? (cdr o))
      (do ((o (cadr o) (car o)))
          ((not (pair? o)) 
           (cond ((symbol? o) (symbol->string o))
                 ((string? o) o)
                 (else #f) ) ) )
      #f ) )


;;; Search a file in a list of directories:
;;; Posix conventions are assumed.

(define (search-file filename path)
  (if (and (> (string-length filename) 0)
           (char=? (string-ref filename 0) #\/) )
      filename
      (let look ((path path))
        (if (pair? path)
            (let ((fullname (string-append (car path) "/" filename)))
              (if (probe-file fullname)
                  fullname
                  (look (cdr path)) ) )
            #f ) ) ) )

;;; Compute the directory part of a pathname:

(define (dirname filename)
  (let loop ((i (- (string-length filename) 1)))
    (if (char=? #\/ (string-ref filename i))
        (substring filename 0 (+ i 1))
        (if (> i 0) 
            (loop (- i 1))
            "./" ) ) ) )
;;; (dirname "/a/b")
;;; (dirname "/a/")
;;; (dirname "/a/lasjdflajs//aaljd")
;;; (dirname "/a.scm")
;;; (dirname "a.scm")

;;; Compute the extension part of a pathname.  Return "" if no
;;; extension is found. The extension does not contain the leading
;;; dot.

(define (extension filename)
  (define ln (string-length filename))
  (let loop ((i (- ln 1)))
    (if (char=? #\. (string-ref filename i))
        (if (< i (- ln 1))
            (substring filename (+ i 1) ln)
            "" )
        (if (> i 0) 
            (loop (- i 1))
            "" ) ) ) )
;;; (extension "/a/b.c")
;;; (extension "/a/b.scm")
;;; (extension "b.scm")
;;; (extension "/t/b.")
;;; (extension "/t/bab")

;;; Define the handler for read errors
;;; Report the line number and the context
(define *read-error-format*
  "~%LiSP2TeX read error (line ~A-~A): ~U ~A~%erroneous line was: ~A " )
(set! *read-error-format* *read-error-format*)

(define (counting-read-error stream start-line message culprits)
  (pp-format stderr-port
             *read-error-format*
             start-line
             (counting-stream-line stream)
             message culprits 
             (counting-stream-current-read-line stream) )
  ;; resume by returning a bizarre object
  '***READ-ERROR-OBJECT*** )

(define (numbering-read-error stream start-line message culprits)
  (pp-format stderr-port
             *read-error-format*
             start-line
             (numbering-stream-line stream)
             message culprits 
             (numbering-stream-current-read-line stream) )
  ;; resume by returning a bizarre object
  '***READ-ERROR-OBJECT*** )

;;; The main function that reads expressions from a counting stream
(define counting-read 'wait)
(define counting-skip 'wait)
(define counting-skip-spaces 'wait)

(let ((fun (make-reader counting-stream-peek-char
                        counting-stream-read-char )))
  (set! counting-read 
        (lambda (stream)
          (call/cc
           (lambda (exit)
             (fun 'skip stream 'useless)
             (let* ((start-line (counting-stream-line stream))
                    (exp (fun 'read stream
                              (lambda (stream message . culprits)
                                (exit
                                 (counting-read-error 
                                  stream start-line message culprits ) ) ) )) )
               exp ) ) ) ) )
  (set! counting-skip-spaces
        (lambda (stream)
          (fun 'skip-spaces stream 'useless) ) )
  (set! counting-skip 
        (lambda (stream)
          (fun 'skip stream 'useless) ) ) )

;;; The main function that reads expressions from a numbering stream
(define numbering-read 'wait)
(define numbering-skip 'wait)

(let ((fun (make-reader 
            numbering-stream-peek-char
            numbering-stream-read-char )) )
  (set! numbering-read 
        (lambda (stream)
          (call/cc 
           (lambda (exit)
             (fun 'skip stream 'useless)
             (let* ((start-line (numbering-stream-line stream))
                    (exp (fun 'read stream
                              (lambda (stream message . culprits)
                                (exit
                                 (numbering-read-error 
                                  stream start-line message culprits ) ) ) )) )
               exp ) ) ) ) )
  (set! numbering-skip 
        (lambda (stream)
          (fun 'skip stream 'useless) ) ) )

;;; end of scan-file.scm
