;;; $Id: scanext.scm,v 1.4 2001/01/06 17:03:51 queinnec Exp $
;;; Copyright (c) 1990-2000 by Christian Queinnec. All rights reserved.
;;; This file is part of LiSP2TeX, a LiSP and TeX merger.
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo
;;;                                                    LiSP2TeX
;;;   Christian Queinnec
;;;   <Christian.Queinnec@lip6.fr> 
;;;   LIP6 --- Universit Paris 6
;;;   4, place Jussieu
;;;   75252 Paris Cedex 05
;;;   France              
;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;;    This program is free software; you can redistribute it and/or modify
;;;    it under the terms of the GNU General Public License as published by
;;;    the Free Software Foundation; either version 2 of the License, or
;;;    (at your option) any later version.
;;;
;;;    This program is distributed in the hope that it will be useful,
;;;    but WITHOUT ANY WARRANTY; without even the implied warranty of
;;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;;    GNU General Public License for more details.
;;;
;;;    You should have received a copy of the GNU General Public License
;;;    along with this program; if not, write to the Free Software Foundation,
;;;    Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

;;;oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo

;;; This file defines the interface with external parsers. When
;;; LiSP2TeX encounters an \ExtractFromFile ("filename" key ...) where
;;; key is a list of names such as ("foo" "bar" "baz"), it looks for
;;; an entity named baz within a container named bar in a container
;;; named foo within "filename". When found the string describing that
;;; entity is extracted from the file and inserted in the output
;;; stream (the insertion is driven by some formats and may be further
;;; transformed).

;;; To scan an external file a process is started whose output is a
;;; sequence of lines with the following format:
;;;      (start real-start end "name1" "name2" ... "nameN")
;;; Each of these line mentions that there exists an entity named nameN 
;;; within ... within name2 within name1. The leading comment is located
;;; between [start ... real-start[ and the body between [real-start end[.

;;; Constraints: If (s rs e "name1" ... "nameN-1" "nameN") appears
;;; then (s' rs' e' "name1" ... "nameN-1") should also appear. There
;;; should be always at least one name.

(define (scan-external-file filename)
  (let* ((ext (extension filename))
         (desc (assoc ext *external-parsers*)) )
    (if desc
        ((cdr desc) filename)
        (command-error *extractfromfile-error6* filename) ) ) )

(define *extractfromfile-error6*
  "~%LiSP2TeX (ExtractFromFile) error: no external parser for ~A~%" )
(set! *extractfromfile-error6* *extractfromfile-error6*)
(define *extractfromfile-error7*
  "~%LiSP2TeX (ExtractFromFile) error: external parser bad status: ~A
Command was: ~A~%" )
(set! *extractfromfile-error7* *extractfromfile-error7*) 
(define *extractfromfile-error8*
  "~%LiSP2TeX (ExtractFromFile) error: lost external parsing result
Command was: ~A~%" )
(set! *extractfromfile-error8* *extractfromfile-error8*)

;;; The Alist of external parsers. They all take two arguments:
;;;    -- the file to parse
;;;    -- the file where to put the parsing results.
;;; The command should be executable of course. The stdout and stderr
;;; of the command are those of LiSP2TeX (pay attention!)
;;; For instance:
;;;    (define-parser "l" "l-parser.sh")

;;; This Alist maps extensions to the function that parses a
;;; file. Predefined parsers exist for Scheme and C files. The
;;; associated function takes a filename and return a Scanned-File
;;; record.

(define *external-parsers* 
  (list (cons "scm" scan-file)
        (cons "c"   scan-c-file)
        (cons "h"   scan-c-file) ) )
(set! *external-parsers* *external-parsers*)

;;; Define for an extension, the parser to use and the name of the
;;; environment to use for extracted snippets.

(define (define-parser extension command)
  (set! *external-parsers* 
        (cons (cons extension (make-scan-external-file command))
              *external-parsers* ) ) )

;;; Wraps an external command into a scanner (a function mapping a
;;; filename into a Scanned-File).

(define (make-scan-external-file command)
  (lambda (filename)
    (let* ((tmp-file (get-tmp-file))
           (cmd (string-append command " " filename " " tmp-file))
           (status (LiSP2TeX-system cmd)) )
      (if (not (= status 0))
          (command-error *extractfromfile-error7* status cmd) )
      (if (not (probe-file tmp-file))
          (command-error *extractfromfile-error8* cmd) )
      (let ((result (read-scanned-expressions filename tmp-file)))
        (delete-file tmp-file)
        result ) ) ) )

;;; The abstract type for externally parsed files.

(define (make-scanned-external-file filename scanned-exps string)
  (vector scan-external-file
          filename scanned-exps string ) )
(define (scanned-external-file-filename sf)
  (vector-ref sf 1) )
(define (scanned-external-file-scanned-expressions sf)
  (vector-ref sf 2) )
(define (scanned-external-file-string sf)
  (vector-ref sf 3) )

;;; Read a file containing parsing results.

(define (read-scanned-expressions orgfile filename)
  (call-with-input-file filename
    (lambda (in)
      (define (check-se o)
        (and (pair? o)
             (integer? (car o))
             (>= (car o) 0)
             (pair? (cdr o))
             (integer? (cadr o))
             (>= (cadr o) (car o))
             (pair? (cddr o))
             (integer? (caddr o))
             (>= (caddr o) (cadr o))
             (pair? (cdddr o))
             (every? string? (cdddr o)) ) )
      (define (accumulate)
        (let ((o (read in)))
          (if (eof-object? o)
              (begin (close-input-port in)
                     '() )
              (if (check-se o)
                  (cons (make-scanned-expression 
                         #f             ; expression
                         'external      ; type
                         (cdddr o)      ; key
                         (car o)        ; start
                         (cadr o)       ; real-start
                         (caddr o) )    ; end
                        (accumulate) )
                  (command-error *extractfromfile-error5* filename o) ) ) ) )
      (make-scanned-external-file
       orgfile
       (list->vector (accumulate))
       (file->string orgfile) ) ) ) )

(define *extractfromfile-error5* 
  "~%LiSP2TeX (ExtractFromFile) error: Incorrect parsing in ~A,
Culprit is ~S~%" )
(set! *extractfromfile-error5* *extractfromfile-error5*)

;;; Find the descriptor of a scanned expression.

(define (find-scanned-external-expression orgkey scanned-file)
  (let* ((ses (scanned-file-scanned-expressions scanned-file))
         (found #f)
         (max (vector-length ses))
         (key (if (pair? orgkey) orgkey (list orgkey))) )
    (define (lookup i)
      (if (< i max)
          (let ((se (vector-ref ses i)))
            (if (compare-keys? key (scanned-expression-key se))
                (begin (set! found #t)
                       (cons se (lookup (+ 1 i))) )
                (lookup (+ 1 i)) ) )
          (if found
              '()
              (begin
                (command-error *find-definition-error* orgkey)
                '(***MISSING-EXPRESSION***) ) ) ) )
    (lookup 0) ) )

;;; Compare a key to find (a list of symbols or strings) with the key of 
;;; a scanned-expression (a symbol or a string or a list of strings).

(define (compare-keys? key names)
  (define (special-equal? k b)
    (cond ((symbol? k) (special-equal? (symbol->string k) b))
          ((symbol? b) (special-equal? k (symbol->string b)))
          (else (equal? k b)) ) )
  (cond ((pair? key)
         (cond ((pair? names)
                (and (special-equal? (car key) (car names))
                     (compare-keys? (cdr key) (cdr names)) ) )
               ((null? names) #f)
               ((and (special-equal? (car key) names)
                     (null? (cdr key)) )
                #t )
               (else #f) ) )
        ((null? key) (null? names))
        (else #f) ) )

;;; end of scanext.scm
